from setuptools import setup, find_packages
from virheat import __version__, _program

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='virheat',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=__version__,
    python_requires=">=3.9",
    license_files=('LICENSE'),
    packages=find_packages(),
    install_requires=[
        "matplotlib>=3.5.1,<=3.8.0",
        "numpy>=1.23.3",
        "pandas>=1.4.4"
    ],
    description='virHEAT creates a heatmap from vcf files and maps positions on to a reference genome.',
    url='https://github.com/jonas-fuchs/virHEAT',
    author='Dr. Jonas Fuchs',
    author_email='jonas.fuchs@uniklinik-freiburg.de',
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)"
    ],
    entry_points="""
    [console_scripts]
    {program} = virheat.command:main
    """.format(program=_program),
    include_package_data=True,
    keywords=[],
    zip_safe=False
)



"""
contains all plotting functions of virHEAT
"""

# BUILT-INS
import math
import numpy as np

# LIBS
import matplotlib.pyplot as plt
import matplotlib.patches as patches


def create_heatmap(ax, frequency_array, cmap):
    """
    create each cell of the heatmap
    """
    # cells
    y_start = 0
    for frequency_list in frequency_array:
        x_start = 0
        for freq in frequency_list:
            heatmap_cell = patches.Rectangle((x_start, y_start), 1, 1, alpha=1, edgecolor="grey", facecolor=cmap.to_rgba(freq))
            ax.add_patch(heatmap_cell)
            x_start += 1
        y_start += 1


def create_genome_vis(ax, genome_y_location, n_mutations, unique_mutations, start, stop):
    """
    create the genome rectangle, mutations and mappings to the heatmap
    """
    # colors
    mutation_type_colors = {
        "SNV": "dimgrey",
        "DEL": "red",
        "INS": "blue"
        }
    mutation_set = set()
    # define coordinates
    y_min = -genome_y_location
    y_max = -genome_y_location+genome_y_location/2

    # create blank rectangle for genome vis
    ax.add_patch(
        patches.FancyBboxPatch(
                            (0, y_min), n_mutations, y_max - y_min,
                            boxstyle="round,pad=-0.0040,rounding_size=0.03",
                            ec="none", fc="lightgrey"
                        )
    )
    # create mutation lines on the genome rectangle and the mapping to the respective cells
    x_start = 0
    length = stop - start
    for mutation in unique_mutations:
        mutation_attributes = mutation.split("_")
        mutation_color = mutation_type_colors[mutation_attributes[3]]
        mutation_set.add(mutation_attributes[3])
        mutation_x_location = n_mutations/length*(int(mutation_attributes[0])-start)
        # create mutation lines
        plt.vlines(x=mutation_x_location, ymin=y_min, ymax=y_max, color=mutation_color)
        # create polygon
        coordinates = [(x_start, 0), (x_start+1, 0), (mutation_x_location, y_max)]
        ax.add_patch(patches.Polygon(coordinates, facecolor=mutation_color, alpha=0.5))
        x_start += 1

    return mutation_set


def create_scores_vis(ax, genome_y_location, n_mutations, n_tracks, unique_mutations, start, stop, score_count, score_name):
    """
    create the scores rectangles, mappings to the reference
    """
    score_set = []
    y_zero = -genome_y_location - (n_tracks + score_count + 1) * (genome_y_location / 2)  # zero line
    length = stop - start

    # create list of tuples [(nt pos, score)]
    for mutation in unique_mutations:
        mutation_attributes = mutation.split("_")
        if not np.isnan(float(mutation_attributes[5])):
            score_set.append((int(mutation_attributes[0]), float(mutation_attributes[5])))
    # check if there is something to plot
    if score_set:
        # define normalization multiplier for the height of the score v lines
        max_value = max([abs(score[1]) for score in score_set])  # max abs in score set
        multiplier = max_value / abs((y_zero + genome_y_location / 4) - y_zero)
        # add text for scale and score_name
        if score_count % 2 == 0:
            score_name_loc, scale_loc, vline_loc = ax.get_xlim()[1] + 0.5, ax.get_xlim()[0] - 0.5, ax.get_xlim()[0]
            name_ha_aln, scale_ha_aln = "left", "right"
        else:
            score_name_loc, scale_loc, vline_loc = ax.get_xlim()[0] - 0.5, ax.get_xlim()[1] + 0.5, ax.get_xlim()[1]
            name_ha_aln, scale_ha_aln = "right", "left"
        # score name
        ax.text(score_name_loc, y_zero, score_name, ha=name_ha_aln, va='center')
        # scale
        ax.text(scale_loc, y_zero, "0", ha=scale_ha_aln, va='center')
        ax.text(scale_loc, y_zero - genome_y_location / 4, -max_value, ha=scale_ha_aln, va='center', color='red')
        ax.text(scale_loc, y_zero + genome_y_location / 4, max_value, ha=scale_ha_aln, va='center', color='blue')
        plt.vlines(x=vline_loc, ymin=y_zero - genome_y_location / 4, ymax=y_zero + genome_y_location / 4, color='black', linewidth=2)
        # add h line boundaries for scale
        if score_count is 1:
            plt.axhline(y=y_zero + genome_y_location / 4, color='grey', linestyle='--', linewidth=0.5)
        plt.axhline(y=y_zero - genome_y_location / 4, color='grey', linestyle='--', linewidth=0.5)
        # create zero line
        plt.axhline(y=y_zero, color='black', linestyle='-', linewidth=0.5)
        # create bars for scores
        for score in score_set:
            # define x value
            mutation_x_location = n_mutations / length * (score[0] - start)
            # create lines for score_set
            if score[1] < 0:
                plt.vlines(x=mutation_x_location, ymin=y_zero + score[1]/multiplier, ymax=y_zero, color="red", linestyle='-')
            else:
                plt.vlines(x=mutation_x_location, ymin=y_zero, ymax=y_zero + score[1] / multiplier, color="blue", linestyle='-')

        return True
    # if not the track is not created
    else:
        print("\033[31m\033[1mERROR:\033[0m Seems like there are no scores in the score set '{}' corresponding to the plotted mutation positions.".format(score_name))
        return False


def create_colorbar(threshold, cmap, min_y_location, n_samples, ax):
    """
    creates a custom colorbar and annotates the threshold
    """
    if n_samples >= 8:
        ticks = [0, 0.2, 0.4, 0.6, 0.8, 1]
        labels = [0, 0.2, 0.4, 0.6, 0.8, 1]
        if threshold + 0.1 in ticks or threshold - 0.1 in ticks:
            rounded_threshold = threshold
        else:
            rounded_threshold = round(threshold * 5) / 5
    else:
        ticks = [0, 0.5, 1]
        labels = [0, 0.5, 1]
        if threshold + 0.25 in ticks or threshold - 0.25 in ticks:
            rounded_threshold = threshold
        else:
            rounded_threshold = round(threshold * 2) / 2

    if rounded_threshold in ticks:
        ticks.remove(rounded_threshold)
        labels.remove(rounded_threshold)
    ticks.append(threshold)
    labels.append(f"threshold\n={threshold}")
    cbar = plt.colorbar(cmap, label="variant frequency", pad=0, shrink=n_samples/(min_y_location+n_samples), anchor=(0.1,1), aspect=15, ax=ax)
    cbar.set_ticks(ticks)
    cbar.set_ticklabels(labels)


def create_mutation_legend(mutation_set, min_y_location, n_samples, n_scores):
    """
    create a legend for the mutation type
    """
    legend_patches = []

    if "DEL" in mutation_set:
        legend_patches.append(patches.Patch(color="red", label='DEL'))
    if "INS" in mutation_set:
        legend_patches.append(patches.Patch(color="blue", label="INS"))
    if "SNV" in mutation_set:
        legend_patches.append(patches.Patch(color="dimgrey", label="SNV"))
    plt.legend(bbox_to_anchor=(1.01, 0.95-(n_samples/(min_y_location+n_samples+n_scores))), handles=legend_patches)


def create_axis(ax, n_mutations, min_y_location, n_samples, file_names, start, stop, genome_y_location, unique_mutations, reference_name):
    """
    create the axis of the plot
    """

    # define plot limits
    ax.set_xlim(0, n_mutations)
    ax.set_ylim(-min_y_location, n_samples)
    # define new ticks depending on the genome size
    axis_length = stop - start
    if n_mutations >= 20:
        xtick_dis = round(axis_length/6, -int(math.log10(axis_length / 6)) + 1)
        xtick_labels = [start, start + xtick_dis, start + xtick_dis*2, start + xtick_dis*3, start + xtick_dis*4, start + xtick_dis*5, stop]
    elif n_mutations >= 10:
        xtick_dis = round(axis_length / 3, -int(math.log10(axis_length / 3)) + 1)
        xtick_labels = [start, start + xtick_dis, start + xtick_dis * 2, stop]
    else:
        xtick_labels = [start, stop]
    xtick_labels = [int(tick) for tick in xtick_labels]
    # get the correct location of the genome pos on the axis
    xticks = [n_mutations/axis_length*(tick - start) for tick in xtick_labels]
    # set new ticks and change spines/yaxis
    ax.set_xticks(xticks, xtick_labels)
    # set y axis labels
    y_ticks = [idx+0.5 for idx in range(0, len(file_names))]
    y_ticks.append(-genome_y_location+genome_y_location/4)
    file_names.append(reference_name)
    ax.set_yticks(y_ticks, file_names)
    # set second x axis for mut pos
    secxtick_labels = []
    secxtick_ticks = []
    for idx, mut in enumerate(unique_mutations):
        mutation = mut.split("_")
        secxtick_labels.append(f"{mutation[1]}{mutation[0]}{mutation[2]}")
        secxtick_ticks.append(idx+0.5)
    secax = ax.secondary_xaxis("top")
    secax.set_xticks(secxtick_ticks, secxtick_labels, rotation=90)
    # set spine visibility
    ax.spines["bottom"].set_position(("data", -genome_y_location))
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)


def create_gene_vis(ax, genes_with_mutations, n_mutations, y_size, n_tracks, start, stop, min_y_location, genome_y_location, colors_genes, n_scores):
    """
    create the vis for the gene
    """

    gene_annotations = []
    mult_factor = n_mutations/(stop-start)

    for idx, gene in enumerate(genes_with_mutations):
        # define the plotting values for the patch
        if genes_with_mutations[gene][0][0] < start:
            x_value = 0
        else:
            x_value = mult_factor*(genes_with_mutations[gene][0][0]-start)
        y_value = -min_y_location+(n_tracks+n_scores-genes_with_mutations[gene][1])*genome_y_location/2
        if genes_with_mutations[gene][0][1] > stop:
            width = n_mutations - x_value
        else:
            width = mult_factor*(genes_with_mutations[gene][0][1]-start) - x_value
        height = genome_y_location/2
        # plot the patch
        ax.add_patch(
            patches.FancyBboxPatch(
                                (x_value, y_value), width, height,
                                boxstyle="round,pad=-0.0040,rounding_size=0.03",
                                ec="black", fc=colors_genes[idx]
                            )
        )
        # define text pos for gene description inside or below the gene box, depending if it fits within
        if width > n_mutations/(y_size*8)*len(gene):
            gene_annotations.append(ax.text(x_value+width/2, y_value+height/2, gene, ha="center", va="center"))
        else:
            gene_annotations.append(ax.text(x_value+width/2, y_value-height/4, gene, rotation=40, rotation_mode="anchor", ha="right", va="bottom"))



"""
contains all data prep functions of virHEAT
"""

# BUILT-INS
import os
import re
import pathlib
import sys

# LIBS
import numpy as np
import pandas as pd


def get_files(path, type):
    """
    returns a list of vcf files in a paticular folder
    """
    return list(pathlib.Path(path).glob(f'*.{type}'))


def get_digit_and_alpha(filename):
    """
    get digits and alpha in file names
    """
    digit_match = re.search(r'\d+', filename)
    if digit_match:
        digit = digit_match.group()
        alpha = re.sub(r'\d+', '', filename)
    else:
        digit = ''
        alpha = filename
    return (alpha, int(digit) if digit else float('inf'))


def convert_string(string):
    """
    converts string to its right type
    """
    string = string.replace("\n", "")
    if string.isdecimal():
        return(int(string))
    elif string.replace('.', '', 1).isdecimal():
        return(float(string))
    else:
        return string


def read_vcf(vcf_file, reference):
    """
    parse vcf files to dictionary
    """
    vcf_dict = {}

    # get header and values
    with open(vcf_file, "r") as f:
        header_lines = [l.strip().split("\t") for l in f if l.startswith('#CHROM')]
        if not header_lines:
            print(f"\033[31m\033[1mWARNING:\033[0m {vcf_file} does not contain a '#CHROM' header!")
            return {}
        header = header_lines[0]
    # get each line as frequency_lists
    with open(vcf_file, "r") as f:
        lines = [l.strip().split("\t") for l in f if l.startswith(reference)]
    # check if vcf is empty
    if not lines:
        print(f"\033[31m\033[1mWARNING:\033[0m {vcf_file} has no variants to {reference}!")
    # get standard headers as keys
    for key in header[0:6]:
        vcf_dict[key] = []
    # functional effect
    vcf_dict["MUT_TYPE_"] = []
    # info field
    for line in lines:
        for info in line[7].split(";"):
            if "=" in info:
                vcf_dict[info.split("=")[0]] = []
    # fill in dictionary
    for line in lines:
        # remember keys that have an entry already
        visited_keys = []
        # check if there are multiple called variants at a single position
        # separated by a comma
        length_variants = len(line[4].split(','))
        for idx, key in enumerate(header[0:6]):
            sublines = line[idx].split(',')
            for i in range(length_variants):
                try:
                    vcf_dict[key].append(convert_string(sublines[i]))
                except IndexError:
                    vcf_dict[key].append(convert_string(sublines[0]))
        # get mutation type
        mutations = line[4].split(',')
        for mutation in mutations:
            if len(line[3]) == len(mutation):
                vcf_dict["MUT_TYPE_"].append("SNV")
            elif len(line[3]) < len(mutation):
                vcf_dict["MUT_TYPE_"].append("INS")
            elif len(line[3]) > len(mutation):
                vcf_dict["MUT_TYPE_"].append("DEL")
        visited_keys.extend(header[0:6])
        visited_keys.append("MUT_TYPE_")
        # get data from info field
        for info in line[7].split(";"):
            if "=" in info:
                key, val = info.split("=")
                val_list = val.split(',')
                for value in val_list:
                    vcf_dict[key].append(convert_string(value))
                visited_keys.append(key)
        # append none for each none visited key in the INFO field
        for key in [k for k in vcf_dict.keys() if k not in visited_keys]:
            vcf_dict[key].extend([None]*length_variants)

    return vcf_dict


def extract_vcf_data(vcf_files, reference, threshold=0, scores=False):
    """
    extract relevant vcf data
    """

    file_names = []
    frequency_lists = []

    for file in vcf_files:
        file_names.append(os.path.splitext(os.path.basename(file))[0])
        vcf_dict = read_vcf(file, reference)
        frequency_list = []
        # write all mutation info in a '_' sep string
        for idx in range(0, len(vcf_dict["#CHROM"])):
            if not vcf_dict["AF"][idx] >= threshold:
                continue
            if scores:
                if vcf_dict['EFF'][idx] is not None:
                    aa_change = vcf_dict['EFF'][idx].split('|')[3]  # extract amino acid changes if provided
                else:
                    aa_change = '-'
                frequency_list.append(
                    (f"{vcf_dict['POS'][idx]}_{vcf_dict['REF'][idx]}_{vcf_dict['ALT'][idx]}_{vcf_dict['MUT_TYPE_'][idx]}_{aa_change}", vcf_dict['AF'][idx])
                )
            else:
                frequency_list.append(
                    (f"{vcf_dict['POS'][idx]}_{vcf_dict['REF'][idx]}_{vcf_dict['ALT'][idx]}_{vcf_dict['MUT_TYPE_'][idx]}", vcf_dict['AF'][idx])
                )

        frequency_lists.append(frequency_list)
    # sort by mutation index
    unique_mutations = sorted(
        {x[0] for li in frequency_lists for x in li}, key=lambda x: int(x.split("_")[0])
    )
    if not unique_mutations:
        sys.exit(f"\033[31m\033[1mERROR:\033[0m No variants to {reference} in all vcf files!")

    return frequency_lists, unique_mutations, file_names


def extract_scores(unique_mutations, scores_file, aa_pos_col, score_col):
    """
    Extract scores from scores_file which corresponding value from aa_pos_col is equal to unique_aa_mutations
    """
    scores_df = pd.read_csv(scores_file)

    # create a dictionary to store the scores for each mutation
    mutation_scores = {}
    for idx, row in scores_df.iterrows():
        mutation_scores[row[aa_pos_col]] = row[score_col]

    unique_scores = []
    for mutation in unique_mutations:
        aa_mut = mutation.split('_')[4]
        if aa_mut in mutation_scores:
            score = mutation_scores[aa_mut]
            unique_scores.append(f"{mutation}_{score}")
        else:
            unique_scores.append(f"{mutation}_nan")

    return unique_scores


def create_freq_array(unique_mutations, frequency_lists):
    """
    create an np array of the mutation frequencies
    """

    frequency_array = []

    for frequency_list in frequency_lists:
        frequencies = []
        for mutation in unique_mutations:
            af = [tup[1] for tup in frequency_list if tup[0] == mutation]
            if af:
                frequencies.append(af[0])
            else:
                frequencies.append(0)
        frequency_array.append(frequencies)

    return np.array(frequency_array)


def annotate_non_covered_regions(coverage_dir, min_coverage, frequency_array, file_names, unique_mutations, reference):
    """
    Insert nan values into np array if position is not covered. Needs
    per base coverage tsv files created by bamqc
    """

    # get tsv files
    per_base_coverage_files = get_files(coverage_dir, "tsv")
    if per_base_coverage_files:
        for i, (file_name, array) in enumerate(zip(file_names, frequency_array)):
            if file_name not in [os.path.splitext(os.path.basename(file))[0] for file in per_base_coverage_files]:
                print(f"\033[31m\033[1mWARNING:\033[0m {file_name} was not found in tsv files.")
                continue
            tsv_file = [file for file in per_base_coverage_files if os.path.splitext(os.path.basename(file))[0] == file_name][0]
            coverage = pd.read_csv(tsv_file, sep="\t")
            coverage = coverage[coverage["#chr"] == reference]
            for j, (mutation, frequency) in enumerate(zip(unique_mutations, array)):
                mut_pos = int(mutation.split("_")[0])
                if coverage[coverage["pos"] == mut_pos].empty or all([frequency == 0, coverage[coverage["pos"] == mut_pos]["coverage"].iloc[0] <= min_coverage]):
                    frequency_array[i][j] = np.NAN

    return np.ma.array(frequency_array, mask=np.isnan(frequency_array))


def delete_common_mutations(frequency_array, unique_mutations):
    """
    delete rows of common mutations (non-zero) that are in the array
    """

    mut_to_del = []

    for idx in range(0, len(frequency_array[0])):
        check_all = []
        for frequency_list in frequency_array:
            check_all.append(frequency_list[idx])
        # check if all mutation in a column are zero (happens with some weird callers)
        if all(x == 0 for x in check_all):
            mut_to_del.append(idx)
        # check if frequencies are present in all columns and the maximal diff is greater than 0.5
        # example [0.8, 0.7, 0.3] is not deleted whereas [0.8, 0.7, 0.7] is deleted
        elif all(x > 0 for x in check_all) and max(check_all)-min(check_all) < 0.5:
            mut_to_del.append(idx)

    for idx in sorted(mut_to_del, reverse=True):
        del unique_mutations[idx]

    return np.delete(frequency_array, mut_to_del, axis=1)


def delete_n_mutations(frequency_array, unique_mutations, min_mut):
    """
    delete mutations that are not present in more than n samples
    """
    mut_to_del = []

    for idx in range(0, len(frequency_array[0])):
        n_mutations = 0
        for frequency_list in frequency_array:
            if frequency_list[idx] > 0:
                n_mutations += 1
        # check if min_mut was reached and if not mark as to delete
        if n_mutations <= min_mut:
            mut_to_del.append(idx)
    # delete the mutations that are found only min_mut times in all samples
    for idx in sorted(mut_to_del, reverse=True):
        del unique_mutations[idx]

    return np.delete(frequency_array, mut_to_del, axis=1)


def zoom_to_genomic_regions(unique_mutations, start_stop):
    """
    restrict the displayed mutations to a user defined genomic range
    """
    zoomed_unique = []

    for mutation in unique_mutations:
        if start_stop[0] <= int(mutation.split("_")[0]) <= start_stop[1]:
            zoomed_unique.append(mutation)

    return zoomed_unique


def parse_gff3(file, reference):
    """
    parse gff3 to dictionary
    """

    gff3_dict = {}

    with open(file, "r") as gff3_file:
        for line in gff3_file:
            # ignore comments and last line
            if not line.startswith(reference):
                continue
            gff_values = line.strip().split("\t")
            # sanity check that the line has a unique ID for the dict key
            # this is a lazy fix as it will exclude e.g. exons without ID and
            # only a parent
            if not gff_values[8].startswith("ID="):
                continue
            # create keys
            if gff_values[2] not in gff3_dict:
                gff3_dict[gff_values[2]] = {}
            # parse the attribute line
            for attribute in gff_values[8].split(";"):
                identifier, val = attribute.split("=")
                # create a new dict for each ID
                if identifier == "ID" and identifier not in gff3_dict:
                    attribute_id = val
                    gff3_dict[gff_values[2]][val] = {}
                # add attributes
                if identifier != "ID":
                    gff3_dict[gff_values[2]][attribute_id][identifier] = val
            # add start, stop and strand
            gff3_dict[gff_values[2]][attribute_id]["start"] = int(gff_values[3])
            gff3_dict[gff_values[2]][attribute_id]["stop"] = int(gff_values[4])

    gff3_file.close()

    if not gff3_dict:
        sys.exit(f"\033[31m\033[1mERROR:\033[0m {reference} not found in gff3 file.")

    return gff3_dict


def get_genome_end(gff3_dict):
    """
    get the end of the genome from the region annotation
    """

    genome_end = 0

    if "region" not in gff3_dict:
        sys.exit("\033[31m\033[1mERROR:\033[0m Region annotation is missing in the gff3!")
    for attribute in gff3_dict["region"].keys():
        stop = gff3_dict["region"][attribute]["stop"]
        if stop > genome_end:
            genome_end = stop

    return genome_end


def create_track_dict(unique_mutations, gff3_info, annotation_type):
    """
    create a dictionary of the genes that have mutations and assess in which
    track these genes should go in case they overlap
    """

    # find genes that have a mutation
    genes_with_mutations = set()

    for mutation in unique_mutations:
        # get the mutation from string
        mutation = int(mutation.split("_")[0])
        for type in annotation_type:
            if type not in gff3_info.keys():
                continue
            for annotation in gff3_info[type]:
                if mutation in range(gff3_info[type][annotation]["start"], gff3_info[type][annotation]["stop"]):
                    if "Name" in gff3_info[type][annotation].keys():
                        attribute_name = gff3_info[type][annotation]["Name"]
                    else:
                        attribute_name = annotation
                    genes_with_mutations.add(
                        (attribute_name,
                         gff3_info[type][annotation]["start"],
                         gff3_info[type][annotation]["stop"])
                    )
    if not genes_with_mutations:
        print("\033[31m\033[1mWARNING:\033[0m either the annotation types were not found in gff3 or the mutations are not within genes.")
        return {}, 0

    # create a dict and sort
    gene_dict = {element[0]: [element[1:4]] for element in genes_with_mutations}
    gene_dict = dict(sorted(gene_dict.items(), key=lambda x: x[1][0]))

    # remember for each track the largest stop
    track_stops = [0]

    for gene in gene_dict:
        track = 0
        # check if a start of a gene is smaller than the stop of the current track
        # -> move to new track
        while gene_dict[gene][0][0] < track_stops[track]:
            track += 1
            # if all prior tracks are potentially causing an overlap
            # create a new track and break
            if len(track_stops) <= track:
                track_stops.append(0)
                break
        # in the current track remember the stop of the current gene
        track_stops[track] = gene_dict[gene][0][1]
        # and indicate the track in the dict
        gene_dict[gene].append(track)

    return gene_dict, len(track_stops)



"""
contains the main workflow of virHEAT
"""
# BUILT-INS
import os
import sys
import math
import argparse

# LIB
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib import colors
from matplotlib import colormaps

# virHEAT
from virheat.scripts import data_prep
from virheat.scripts import plotting
from virheat import __version__
from . import _program


def get_args(sysargs):
    """
    arg parsing for virheat
    """
    parser = argparse.ArgumentParser(
        prog=_program,
        usage='''\tvirheat <folder containing input files (vcf/tsv)> <output dir> -r -l/-g [additional arguments]''')
    parser.add_argument(
        "input",
        nargs=2,
        help="folder containing input files and output folder"
    )
    parser.add_argument(
        "-r",
        "--reference",
        type=str,
        metavar="ref_id",
        required=True,
        default=None,
        help="reference identifier"
    )
    parser.add_argument(
        "--name",
        type=str,
        metavar="virHEAT_plot.pdf",
        default="virHEAT_plot.pdf",
        help="plot name and file type (pdf, png, svg, jpg). Default: virHEAT_plot.pdf"
    )
    parser.add_argument(
        "-l",
        "--genome-length",
        type=int,
        metavar="None",
        default=None,
        help="length of the genome (needed if gff3 is not provided)"
    )
    parser.add_argument(
        "-g",
        "--gff3-path",
        type=str,
        metavar="None",
        default=None,
        help="path to gff3 (needed if length is not provided)"
    )
    parser.add_argument(
        "-a",
        "--gff3-annotations",
        type=str,
        action="store",
        metavar="gene",
        nargs="*",
        default=["gene"],
        help="annotations to display from gff3 file (standard: gene). Multiple possible."
    )
    parser.add_argument(
        "-t",
        "--threshold",
        type=float,
        metavar="0",
        default=0,
        help="display frequencies above this threshold (0-1)"
    )
    parser.add_argument(
        "--delete",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="delete mutations that are present in all samples and their maximum frequency divergence is smaller than 0.5"
    )
    parser.add_argument(
        "-n",
        "--delete-n",
        type=int,
        metavar='None',
        default=None,
        help="do not show mutations that occur n times or less (default: Do not delete)"
    )
    parser.add_argument(
        "-z",
        "--zoom",
        type=int,
        metavar=("start", "stop"),
        nargs=2,
        help="restrict the plot to a specific genomic region."
    )
    parser.add_argument(
        "--sort",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="sort sample names alphanumerically"
    )
    parser.add_argument(
        "--min-cov",
        type=int,
        metavar="20",
        default=20,
        help="display mutations covered at least x time (only if per base cov tsv files are provided)"
    )
    parser.add_argument(
        "-s", "--scores",
        metavar=('scores_file', 'pos_col', 'score_col', 'score_name'),
        nargs=4,
        action='append',
        help="specify scores to be added to the plot by providing a CSV file containing scores, along with its column for amino-acid positions, its column for scores, and descriptive score names (e.g., expression, binding, antibody escape, etc.). This option can be used multiple times to include multiple sets of scores."
    )
    parser.add_argument(
        "-v",
        "--version",
        action='version',
        version=f"virheat {__version__}"
    )
    if len(sysargs) < 1:
        parser.print_help()
        sys.exit(-1)
    else:
        return parser.parse_args(sysargs)


def main(sysargs=sys.argv[1:]):
    """
    main function for data extraction and plotting
    """

    # parse args
    args = get_args(sysargs)

    # get vcf files and sort
    vcf_files = data_prep.get_files(args.input[0], "vcf")
    if args.sort:
        vcf_files = sorted(vcf_files, key=lambda x: data_prep.get_digit_and_alpha(os.path.basename(x)))

    # extract vcf info
    n_scores = 0
    if not vcf_files:
        sys.exit("\033[31m\033[1mERROR:\033[0m No VCF files provided")
    else:
        if args.scores:
            frequency_lists, unique_mutations, file_names = data_prep.extract_vcf_data(vcf_files, args.reference, threshold=args.threshold, scores=True)
            n_scores = len(args.scores)
        else:
            frequency_lists, unique_mutations, file_names = data_prep.extract_vcf_data(vcf_files, args.reference, threshold=args.threshold)

    if args.zoom:
        unique_mutations = data_prep.zoom_to_genomic_regions(unique_mutations, args.zoom)
    frequency_array = data_prep.create_freq_array(unique_mutations, frequency_lists)

    # enables the deletion option only if more than 1 vcf file is provided
    if len(vcf_files) > 1:
        # user specified delete options (removes mutations based on various rationales)
        if args.delete:
            frequency_array = data_prep.delete_common_mutations(frequency_array, unique_mutations)
        if args.delete_n is not None:
            frequency_array = data_prep.delete_n_mutations(frequency_array, unique_mutations, args.delete_n)

    # annotate low coverage if per base coverage from qualimap was provided
    data_prep.annotate_non_covered_regions(args.input[0], args.min_cov, frequency_array, file_names, unique_mutations, args.reference)

    # define relative locations of all items in the plot
    n_samples, n_mutations = len(frequency_array), len(frequency_array[0])
    if n_mutations == 0:
        sys.exit("\033[31m\033[1mERROR:\033[0m Frequency array seems to be empty. There is nothing to plot.")
    if n_samples < 4:
        genome_y_location = 2
    else:
        genome_y_location = math.log2(n_samples)

    # gff3 data extraction and define y coordinates
    if args.gff3_path is not None and args.genome_length is not None:
        sys.exit("\033[31m\033[1mERROR:\033[0m Do not provide the -g and -l argument simultaneously!")
    elif args.gff3_path is not None:
        gff3_info = data_prep.parse_gff3(args.gff3_path, args.reference)
        genome_end = data_prep.get_genome_end(gff3_info)
        genes_with_mutations, n_tracks = data_prep.create_track_dict(unique_mutations, gff3_info, args.gff3_annotations)
    elif args.genome_length is not None:
        genome_end = args.genome_length
        n_tracks = 0
    else:
        sys.exit("\033[31m\033[1mERROR:\033[0m Provide either a gff3 file (-g) or the length (-l) of the genome which you used for mapping")

    # define min y coordinate
    if n_tracks != 0 or n_scores != 0:
        min_y_location = genome_y_location + genome_y_location / 2 * (n_tracks + n_scores + 2)
    else:
        min_y_location = genome_y_location

    # define size of the plot
    y_size = n_mutations*0.4
    if args.scores:
        x_size = y_size * (n_samples + min_y_location + n_scores) / n_mutations
    else:
        x_size = y_size*(n_samples + min_y_location)/n_mutations
    x_size = x_size-x_size*0.15  # compensate of heatmap annotation

    # ini the fig
    fig, ax = plt.subplots(figsize=[y_size, x_size])

    # define boundaries for the plot
    if args.zoom:
        start, stop = args.zoom[0], args.zoom[1]
        # rescue plot if invalid zoom values are given
        if args.zoom[0] < 0:
            start = 0
        if args.zoom[1] > genome_end:
            stop = genome_end
    else:
        start, stop = 0, genome_end

    # plot all common elements
    cmap = cm.gist_heat_r
    cmap.set_bad('silver', 1.)
    cmap_cells = cm.ScalarMappable(norm=colors.Normalize(0, 1), cmap=cmap)
    plotting.create_heatmap(ax, frequency_array, cmap_cells)
    mutation_set = plotting.create_genome_vis(ax, genome_y_location, n_mutations, unique_mutations, start, stop)
    plotting.create_axis(ax, n_mutations, min_y_location, n_samples, file_names, start, stop, genome_y_location,
                         unique_mutations, args.reference)
    plotting.create_mutation_legend(mutation_set, min_y_location, n_samples, n_scores)
    plotting.create_colorbar(args.threshold, cmap_cells, min_y_location, n_samples, ax)
    # plot gene track
    if args.gff3_path is not None:
        if genes_with_mutations:
            # distinct colors for the genes
            cmap_genes = plt.get_cmap('tab20', len(genes_with_mutations))
            colors_genes = [cmap_genes(i) for i in range(len(genes_with_mutations))]
            # plot gene track
            plotting.create_gene_vis(ax, genes_with_mutations, n_mutations, y_size, n_tracks, start, stop, min_y_location, genome_y_location, colors_genes, n_scores)
    # plot scores as track below the gene/genome track
    if args.scores:
        score_count = 1
        for score_params in args.scores:
            scores_file, pos_col, score_col, score_name = score_params
            unique_scores = data_prep.extract_scores(unique_mutations, scores_file, pos_col, score_col)
            track_created = plotting.create_scores_vis(ax, genome_y_location, n_mutations, n_tracks, unique_scores, start, stop, score_count, score_name)
            if track_created:  # only creates a track if it finds mutations to annotate
                score_count += 1
    # create output folder
    if not os.path.exists(args.input[1]):
        os.makedirs(args.input[1])

    # save fig
    fig.savefig(os.path.join(args.input[1], args.name), bbox_inches="tight")

